/*Copyright (c) 2012 Moussa Dembélé (www.mousman.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com.mousman.as3.agal.parser.objs 
{
	import com.mousman.as3.agal.parser.dto.MBuffersWriterDTO;
	import com.mousman.as3.agal.parser.shaders.ShaderStore;
	import com.mousman.as3.structure.garbage.IDestroyable;
	import flash.display3D.Context3D;
	import flash.display3D.IndexBuffer3D;
	import flash.display3D.VertexBuffer3D;
	import flash.utils.ByteArray;
	import mx.core.ByteArrayAsset;
	/**
	 * import a mdm binary file and create a set of vertex buffers and index buffers 
	 * @author Moussa Dembélé (www.mousman.com) based on Christer Kaitila and Nathan Ostgard work
	 * 
	 */
	public class MObjImportParser implements IDestroyable
	{
		// the context3D in wich we want to upload the buffers
		private var _context3D:Context3D;		
		
		// contains all the datas for the mesh
		/**
		 * @private
		 */		
		private var _mesh:MObjMesh;
		
		// contains colors that we can force in the construction of the mesh
		private var _writerParams:MBuffersWriterDTO;
		// contains all the shaders we want to use
		private var _shaderStore:ShaderStore;
		
		private var _injectScreenTexture:Boolean;
		private var _highestLowest:Object;
		
		/**
		* Constructor. Creates a new MObjParser instance.
		*
		* <p>After you create the MObjParser instance, call the parse method to create the mesh datas.</p>
		* 
		* @param context : Context3D.<p>Used to create and upload VertexBuffers and IndexBuffers</p>
		* @param shaderStore : ShaderStore.<p>Contains all the shaders used for the mesh</p>
		*/		
		public function MObjImportParser(context:Context3D, shaderStore:ShaderStore) {
			_context3D = context;
			_shaderStore = shaderStore;
		}
		//----------------------------------------------------------------------------------------------
		/**
		* Parse the obj and mtl file and create the mesh datas
		*
		* 
		* @param meshData : ByteArrayAsset.<p>Contains the obj and mtl datas in a binary format</p>
		* @param writerParams : MBuffersWriterDTO <p>used to force color on the mesh.</p>
		* 
		* @return the datas of the mesh
		*/			
		//----------------------------------------------------------------------------------------------		
		public function parse(meshData:ByteArrayAsset, writerParams:MBuffersWriterDTO = null,injectScreenTexture:Boolean = false): MObjMesh {
			_mesh = new MObjMesh();
			if(_writerParams == null)
				_writerParams = new MBuffersWriterDTO();
			_injectScreenTexture = injectScreenTexture;
			// read and explode the datas
			parseObj(meshData);
			return _mesh;
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function parseObj(meshData:ByteArrayAsset):void {
			var buffer3D:VertexBuffer3D;
			var index3D:IndexBuffer3D;
			var object3Dset:MObject3D;
			var mPolygonsGroup:MPolygoneGroup3D;
			var totalVertices:int;
			var startObjPos:uint;
			var endObjPos:uint;
			var size:uint;
			
			readHeader(meshData);
			var nbObj:uint = meshData.readUnsignedInt();
			//log("nbObj : " + nbObj);
			
			for (var i:int = 0; i < nbObj; i++) {
				var object3D:MObject3D = readObj(meshData);
				_mesh.add(object3D);
			}
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readObj(meshData:ByteArrayAsset):MObject3D 
		{
			var objSize:uint = meshData.readUnsignedInt();
			var id:String = meshData.readUTF();
			//log("-----------------------");
			//log("MObject3D : " + id);
			var object3D:MObject3D = new MObject3D(id);
			var nbPGroups:uint = meshData.readUnsignedInt();
			//log("nbPGroups : " + nbPGroups);
			var polygonsGroups:Vector.<MPolygoneGroup3D> = new Vector.<MPolygoneGroup3D>(nbPGroups,true);
			object3D.polygonsGroups = polygonsGroups;
			
			for (var i:int = 0; i < nbPGroups; i++) {
				var polygoneGroup3D:MPolygoneGroup3D = readPGroup(meshData);
				polygonsGroups[i] = polygoneGroup3D;
			}
			
			return object3D;
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readPGroup(meshData:ByteArrayAsset):MPolygoneGroup3D 
		{
			var groupSize:uint = meshData.readUnsignedInt();
			var id:String = meshData.readUTF();
			log("   MPolygoneGroup3D : " + id);
			
			var polygoneGroup3D:MPolygoneGroup3D;
			polygoneGroup3D = new MPolygoneGroup3D(id);
			
			var nbVertices:uint = meshData.readUnsignedInt();
			polygoneGroup3D.nbVertices = nbVertices;
			var smoothShading:Number = meshData.readDouble();
			polygoneGroup3D.smoothShading = smoothShading;
			
			var mtlSize:uint = meshData.readUnsignedInt();
			var mtlId:String = meshData.readUTF();
			var mtlSet:MMtlSet = new MMtlSet(mtlId);
			polygoneGroup3D.mtlSet = mtlSet;
			mtlSet.diffuseColor = meshData.readUnsignedInt();
			mtlSet.ambiantColor = meshData.readUnsignedInt();
			mtlSet.specularColor = meshData.readUnsignedInt();
			mtlSet.specularCoef= meshData.readDouble();
			mtlSet.emissiveColor = meshData.readUnsignedInt();
			mtlSet.dissolved = meshData.readDouble();
			mtlSet.illumination = meshData.readDouble();
			var aTexture:String = meshData.readUTF();
			if (aTexture !="")
				mtlSet.ambiantTexture = aTexture;
			var dTexture:String = meshData.readUTF();
			if (dTexture !="")
				mtlSet.diffuseTexture = dTexture;
			var isTexture:Boolean = meshData.readBoolean();
			var nbBufferSets:uint = meshData.readUnsignedInt();
			for (var i:int = 0; i < nbBufferSets; i++) {
				var bufferSet:MPolygonGroupBuffers = readBufferSet(meshData);
				polygoneGroup3D.addPolygonBuffers(bufferSet);
			}		
		
			return polygoneGroup3D;			
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readBufferSet(meshData:ByteArrayAsset):MPolygonGroupBuffers 
		{
			var bSize:uint = meshData.readUnsignedInt();
			var id:String = meshData.readUTF();
			//log("       MPolygonGroupBuffers : " + id);
			var bufferSet:MPolygonGroupBuffers = new MPolygonGroupBuffers(id);
			var dataPerVertex:int = meshData.readInt();
			var nbVertices:int = meshData.readInt();
			var verticesBA:ByteArray = new ByteArray();
			//log("       nbVertices : " + nbVertices + " , dataPerVertex: " + dataPerVertex);
			meshData.readBytes(verticesBA, 0, nbVertices * dataPerVertex * 4);
			var nbIndices:int = meshData.readInt();
			//log("       nbIndices : " + nbIndices );
			var indices:Vector.<uint> = new Vector.<uint>();
			for (var i:int = 0; i < nbIndices; i++) {
				indices.push(meshData.readUnsignedInt());
			}		
			
			var vertexBuffer:VertexBuffer3D = _context3D.createVertexBuffer(nbVertices, dataPerVertex);
			//log("creating " + nbVertices +" vertex and index");
			vertexBuffer.uploadFromByteArray(verticesBA,0, 0, nbVertices);
			bufferSet.vertexBuffer = vertexBuffer;
			
			// index
			var indexBuffer:IndexBuffer3D = _context3D.createIndexBuffer(nbIndices);
			indexBuffer.uploadFromVector(indices, 0, nbIndices);
			bufferSet.indexBuffer = indexBuffer;
			
			return bufferSet;
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function readHeader(meshData:ByteArrayAsset):void 
		{
			var compressTypes:Array = ["no compression", "lzma", "zlib"];
			var compressModes:Array = ["not compact", "compact"];
			
			var signature:String = meshData.readUTFBytes(3);
			
			var compressType:int = meshData.readByte();
			log("File compression : " + compressTypes[compressType]);
			var compactMode:int = meshData.readByte();
			log("Compact mode  : " + compressModes[compactMode]);
			
			//if (compactMode != 0) {
				//log("Wrong compact mode !");
				//throw new Error("Wrong compact mode !");
			//}
			
			if(_injectScreenTexture){
				meshData.position = 78;
				_highestLowest = new Object();
				_highestLowest.highestX = meshData.readFloat();
				_highestLowest.lowestX = meshData.readFloat();
				_highestLowest.highestY = meshData.readFloat();
				_highestLowest.lowestY = meshData.readFloat();
				_highestLowest.highestZ = meshData.readFloat();
				_highestLowest.lowestZ = meshData.readFloat();
			}
			else meshData.position = 102;
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function log(txt:String):void {
			trace(txt);
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function clean():void {
		}
		//----------------------------------------------------------------------------------------------	
		/**
		* delete datas
		*/				
		//----------------------------------------------------------------------------------------------	
		public function destroy():void {
			_mesh.destroy();
			_mesh = null;
			_writerParams = null;
			_shaderStore.destroy();
			_shaderStore = null;
			_context3D = null;
		}
		//----------------------------------------------------------------------------------------------	
		/**
		* the object containing mesh datas
		*/			
		//----------------------------------------------------------------------------------------------	
		/**
		* the object containing mesh datas
		*/			
		public function get mesh():MObjMesh {return _mesh;}
		
		public function get highestLowest():Object {return _highestLowest;}
		public function set highestLowest(value:Object):void {_highestLowest = value;}
	}

}
