/*Copyright (c) 2012 Moussa Dembélé (www.mousman.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package 
{
	import com.adobe.utils.AGALMiniAssembler;
	import com.mousman.as3.agal.shader.IMShader;
	import flash.display3D.Context3D;
	import flash.display3D.Context3DProgramType;
	import flash.display3D.Context3DVertexBufferFormat;
	import flash.display3D.Program3D;
	import flash.display3D.VertexBuffer3D;
	/**
	 * @author Moussa Dembélé (www.mousman.com)
	 * va0 : position x,y,z
	 * va1 : color r,g,b
	 * va2 : normales
	 * 
	 * vc0 : scene matrix
	 * vc4 : object matrix
	 * vc8 : object matrix without translation (only rotation)
	 * vc12 : x : flat value 
	 * vc13 : texture positionment matrix
	 * vc17 : light position
	 * vc21 : camera position
	 * 
	 * fc0 : ambient light
	 * fc1 : light diffuse color
	 * fc2 : light position
	 * fc3 : light specular color
	 * fc4 : camera position
	 * fc5 : material ambiant color
	 * fc6 : specular material color	
	 */
	public class MPhongFlatShader implements IMShader
	{
		
		private var _vertexProgram:String;
		private var _fragmentProgram:String;
		private var _shaderProgram:Program3D;
		private var _vertexShaderAssembler:AGALMiniAssembler;
		private var _fragmentShaderAssembler:AGALMiniAssembler;			
		private var _dataPerVertex:int = 7;
		
		public function MPhongFlatShader() 
		{
			createVertexProgram();
			createFragmentProgram();			
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function createVertexProgram():void {
			_vertexProgram =	
				"mov vt0        ,va0                \n" + // x,y,z coordinates
				"m44 vt0        ,vt0     ,vc4       \n" + // apply model matrix
			    "div vt0.z      ,vt0.z   ,vc12.x    \n" + // flattening model
				"m44 vt0        ,vt0     ,vc13      \n" + // texture linked positionment matrix
				"m44 vt0        ,vt0     ,vc0       \n" + // transform vertex x,y,z
				"mov op         ,vt0                \n" + // 
				"mov v0         ,va1                \n" + // move vertex r,g,b to fragment shader
				"m44 v1         ,va2     ,vc8       \n" + // v1, interpolate normal position in the scene
				
				"m44 vt1        ,va0     ,vc4       \n" + // interpolate object position in the scene	
				"sub vt2        ,vc17    ,vt1       \n" + // get dir from the fragment to the light
				"mov v2         ,vt2                \n" + // 
				
				"sub vt3        ,vc21    ,vt1       \n" + // get dir from the fragment to camera
				"mov v3         ,vt3                \n";  // 
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function createFragmentProgram():void {
			_fragmentProgram = 
				"nrm ft0.xyz,   v2.xyz       \n" +  // normalize the light direction	
				"mov ft0.w,     fc2.w        \n" +  
				
				"nrm ft1.xyz   ,v3            \n" + // normalize the camera direction	
				"mov ft1.w     ,fc2.w         \n" + 	
				
				"mov ft2       ,v1            \n" + //normals
				"nrm ft2.xyz   ,ft2           \n" +
				"mov ft2.w     ,fc2.w         \n" + // 
				
				"mov ft3       ,v0            \n" + // move vertex r,g,b 
				
				 //calculate diffuse term - ft4
				"dp3 ft4,       ft2,       ft0          \n" + // find projection of direction to light on normal - ft4
				"mov ft5,       ft4                     \n" + // save for specular light
				"sat ft4,       ft4        	   	        \n" + // ignore negative values
				"mul ft4.xyz,   ft4.xyz,   fc1.w        \n" + // multiply projection with light's diffuse amount
				"mul ft4.xyz,   ft4.xyz    fc1.xyz      \n" + // multiply projection of direction to light on normal with light's diffuse color
				"mul ft4.xyz,   ft4.xyz,   ft3          \n" + // multiply by material's diffuse color

				 //calculate specular term - ft5
				 //reflected light
				"mul ft5,      ft5,        ft2       \n" + 
				"add ft5,      ft5,        ft5       \n" + 
				
				"sub ft5,      ft5,        ft0       \n" + 
				 
				"dp3 ft5        ft5,       ft1       \n" + 
				"sat ft5,       ft5                  \n" + // 
				"pow ft5,       ft5,       fc6.w     \n" + // apply concentration (gloss)
				"mul ft5.xyz,   ft5.xyz,   fc3.xyz   \n" + // multiply with light's specular color
				"mul ft5.xyz,   ft5.xyz,   fc3.w     \n" + // multiply with light's specular amount
				"mul ft5.xyz,   ft5.xyz,   fc6.xyz   \n" + // multiply with material's specular color
				
				"add ft6.xyz,   ft5.xyz,   ft4.xyz   \n" + // combine diffuse and specular terms
				
				"mov ft7, fc0                        \n" + // ambiant light
				"mul ft7.xyz,   ft7.xyz,   fc5.xyz   \n" + // multiply   ambiant light color with  material's ambiant color
				"mul ft7.xyz,   ft7.xyz,   fc0.w     \n" + // multiply with ambiant light amount
				
				"add ft7.xyz,   ft7.xyz,   ft6.xyz   \n" + // add diffuse and specular terms 
				
				"mov ft7.w,      v0.w             \n" + // alpha value
				"mov oc,         ft7              \n";  // output		
		}	
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function setActiveVertextBuffer(context3D:Context3D, vertexBuffer:VertexBuffer3D):void {
			// XYZ
			context3D.setVertexBufferAt(0, vertexBuffer, 0, Context3DVertexBufferFormat.FLOAT_3);
			// colors
			context3D.setVertexBufferAt(1, vertexBuffer, 3, Context3DVertexBufferFormat.BYTES_4);
			// normales
			context3D.setVertexBufferAt(2, vertexBuffer, 4, Context3DVertexBufferFormat.FLOAT_3);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function releaseActiveVertextBuffer(context3D:Context3D):void {
			// XYZ
			context3D.setVertexBufferAt(0, null);
			// colors
			context3D.setVertexBufferAt(1, null);
			// normales
			context3D.setVertexBufferAt(2, null);
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function assembleNupload(context3D:Context3D):void {
			 assemble(context3D);
			 upload();
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function assemble(context3D:Context3D):void {
			_vertexShaderAssembler = new AGALMiniAssembler();
			_vertexShaderAssembler.assemble(Context3DProgramType.VERTEX, _vertexProgram);
			
			_fragmentShaderAssembler = new AGALMiniAssembler();
			_fragmentShaderAssembler.assemble(Context3DProgramType.FRAGMENT, _fragmentProgram);
			_shaderProgram = context3D.createProgram();
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function upload():void {
			_shaderProgram.upload(_vertexShaderAssembler.agalcode, _fragmentShaderAssembler.agalcode);
		}	
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function destroy():void {
			_shaderProgram.dispose();
		}			
		//----------------------------------------------------------------------------------------------		
		public function get vertexProgram():String{return _vertexProgram;}
		public function get fragmentProgram():String { return _fragmentProgram; }
		public function get shaderProgram():Program3D {return _shaderProgram;}
		
		public function get dataPerVertex():int {return _dataPerVertex;}
	}

}