/*Copyright (c) 2012 Moussa Dembélé (www.mousman.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package 
{
	import com.mousman.as3.agal.light.IMLight;
	import com.mousman.as3.agal.mesh.IMMesh;
	import com.mousman.as3.agal.parser.objs.MMtlSet;
	import com.mousman.as3.agal.parser.objs.MObject3D;
	import com.mousman.as3.agal.parser.objs.MObjImportParser;
	import com.mousman.as3.agal.parser.objs.MObjMesh;
	import com.mousman.as3.agal.parser.objs.MPolygoneGroup3D;
	import com.mousman.as3.agal.parser.objs.MPolygonGroupBuffers;
	import com.mousman.as3.agal.parser.shaders.ShaderSet;
	import com.mousman.as3.agal.parser.shaders.ShaderStore;
	import com.mousman.as3.agal.parser.verticeswriter.MVerticesBytesArrayWriter;
	import com.mousman.as3.agal.shader.ShaderType;
	import com.mousman.as3.utils.color.ColorUtil;
	import flash.display3D.Context3D;
	import flash.display3D.Context3DProgramType;
	import flash.display3D.IndexBuffer3D;
	import flash.display3D.Program3D;
	import flash.display3D.textures.Texture;
	import flash.display3D.VertexBuffer3D;
	import flash.geom.Matrix3D;
	import flash.geom.Vector3D;
	import flash.utils.CompressionAlgorithm;
	import mx.core.ByteArrayAsset;
	
	/**
	 * @author Moussa Dembélé (www.mousman.com)
	 */
	public class PlanMesh implements IMMesh
	{
		[Embed(source="../lib/plan.mdm", mimeType="application/octet-stream")]
		private var _meshData:Class;
		private var _meshBA:ByteArrayAsset;
		
		// contains all the ShaderNCo
		private var _shaderStore:ShaderStore;
		
		private var _context3D:Context3D;
		private var _currentProgram:Program3D;
		
		private var _parser:MObjImportParser;
		private var _mesh:MObjMesh;
		private var _modelMatrix:Matrix3D;
		private var _modelPosition:Vector3D;
		private var _modelRotation:Vector3D;
		
		private var _modelViewProjection:Matrix3D;
		
		private var _lights:Vector.<IMLight>;
		private var _cameraMatrix:Matrix3D;
		private var _ambiantLight:Vector.<Number>;
		
		private var _xRange:Number;
		private var _xOffset:Number;
		private var _yRange:Number;
		private var _yOffset:Number;
		private var _zRange:Number;
		private var _zOffset:Number;
		
		private var _screenTexture:Texture;
		private var _xTextureScale:Number =1;
		private var _yTextureScale:Number =1;		
		
		// store the textures
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function PlanMesh(context3D:Context3D){
			
			_context3D = context3D;
			_modelMatrix = new Matrix3D();
			_modelPosition = new Vector3D();
			_modelRotation = new Vector3D();
			_lights = new Vector.<IMLight>();
			
			// definition and decalration of the shaders to use
			setShaders();
			
			// parse the OBJ file and create buffers
			_meshBA = ByteArrayAsset(new _meshData()); 
			_meshBA.uncompress(CompressionAlgorithm.LZMA);
			_parser = new MObjImportParser(_context3D, _shaderStore);
			_mesh = _parser.parse(_meshBA,null,true);
			
			// assemble and upload program
			initPrograms();
			
			// show vertices number
			//count();
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------
		private function count():void{
			var nbIndex:int = 0;
			var iBuffer:int = 0;
			var buffer3D:VertexBuffer3D;
			var index3D:IndexBuffer3D;
			var object3Dset:MObject3D;
			var mPolygonsGroup:MPolygoneGroup3D;
			var polygonBufferSet :MPolygonGroupBuffers;
			var totalVertices:int;
			
			for each(object3Dset in _mesh.mObjects) {
				//trace("--------  object " + object3Dset.id);
				var totalObjectVertices:int =0;
				for each(mPolygonsGroup in object3Dset.polygonsGroups) {
					//trace("   polygon group : " + mPolygonsGroup.id + "    nbVertices : " +mPolygonsGroup.nbVertices);
					//trace("      use texture : " + mPolygonsGroup.mtlSet.ambiantTexture + " , " + mPolygonsGroup.mtlSet.diffuseTexture);
					totalObjectVertices += mPolygonsGroup.nbVertices;
					for each(polygonBufferSet in mPolygonsGroup.polygonBufferSet) {
						//trace("MPolygonGroupBuffers");
					}
					
				}
				totalVertices += totalObjectVertices;
				//trace("total vertices for object " + object3Dset.id +" : " + totalObjectVertices);
			}
			//trace("total vertices : " + totalVertices);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		private function setShaders():void {
			
			// ShaderSets's container
			_shaderStore = new ShaderStore();
			
			// definition and decalration of the shaders to use
			var shader:MSimpleShaderScreenTex = new MSimpleShaderScreenTex();
			var writer:MVerticesBytesArrayWriter = new MVerticesBytesArrayWriter();
			
			// a ShaderSet object contains a IMShader and a IMBytesArrayWriter
			var shaderSet:ShaderSet = new ShaderSet(ShaderType.DEFAULT_SHADER, shader, writer);
			_shaderStore.add(shaderSet);

		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		private function initPrograms():void {
			// create and upload programs
			var shaderSet:ShaderSet;
			
			shaderSet = _shaderStore.shaderSetArray[ShaderType.DEFAULT_SHADER];
			shaderSet.assembleNupload(_context3D);
			
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function addLight(light:IMLight):void {
			_lights.push(light);
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function setSaticConstants():void {
			var color:Object;
			var pos:Vector.<Number>;
			//fc0, ambient light 
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT, 0, _ambiantLight);
			//fc1, light diffuse color
			color = ColorUtil.hexArgb(_lights[0].diffuseColor);
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT, 1, Vector.<Number>([Number(color.rouge)/255,Number(color.vert)/255,Number(color.bleu)/255,_lights[0].diffuseIntensity]));
			//fc2, light position	
			pos =  Vector.<Number>([_lights[0].position.x,_lights[0].position.y,_lights[0].position.z,1]);
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT, 2, pos);		
			//fc4, values for u,v mapping
			pos =  Vector.<Number>([1, 2, 0, 0]);
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT, 4, pos);				
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		private function setFrameConstants():void {
			var pos:Vector.<Number>;
			//vc0, scene
			_context3D.setProgramConstantsFromMatrix(Context3DProgramType.VERTEX, 0, _modelViewProjection, true);
			//vc4, object position
			_context3D.setProgramConstantsFromMatrix(Context3DProgramType.VERTEX, 4, _modelMatrix, true); 
			//vc8, object rotation
			var rotMatrix:Matrix3D = _modelMatrix.clone();
			rotMatrix.transpose();
			rotMatrix.invert();			
			_context3D.setProgramConstantsFromMatrix(Context3DProgramType.VERTEX, 8, rotMatrix, true);	
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function drawTriangles():void {
			var shaderSet:ShaderSet;
			var buffer3D:VertexBuffer3D;
			var index3D:IndexBuffer3D;
			var object3Dset:MObject3D;
			var mPolygonsGroup:MPolygoneGroup3D;
			var polygonBufferSet:MPolygonGroupBuffers;
			var color:Object;
			var specCoef:Number;
			
			// ----------------------
			// define the constantes wich are not dependant of groups
			// ----------------------			
			setFrameConstants();
			shaderSet = _shaderStore.getShaderById(ShaderType.DEFAULT_SHADER);
			_context3D.setProgram(shaderSet.program);
			_context3D.setTextureAt(0, _screenTexture);
			
			
			// loop through the objects
			var oSize:int = _mesh.mObjects.length;
			for (var i:int = 0; i < oSize; i++ ) {
				object3Dset =  _mesh.mObjects[i];
				//trace("drawing object " + object3Dset.id);
				
				// loop through the groups
				var pgSize:int = object3Dset.polygonsGroups.length;
				for (var j:int = 0; j < pgSize; j++ ) {
					mPolygonsGroup = object3Dset.polygonsGroups[j];
					//trace("drawing group " + mPolygonsGroup.id);
					// define the constantes depending on groups
					setDefaultConstant(mPolygonsGroup.mtlSet);
					
					// loop through the buffers
					var pbSize:int = mPolygonsGroup.polygonBufferSet.length;
					for (var k:int = 0; k < pbSize; k++ ) {
						polygonBufferSet = mPolygonsGroup.polygonBufferSet[k];
						buffer3D = polygonBufferSet.vertexBuffer;
						index3D = polygonBufferSet.indexBuffer;
						shaderSet.shader.setActiveVertextBuffer(_context3D, buffer3D)
						//trace("drawing triangles ");
						_context3D.drawTriangles(index3D, 0, -1);
					}
				}
			}
			
			_context3D.setTextureAt(0, null);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function setDefaultConstant(mtlSet:MMtlSet):void {
			// fc3 , material ambiant color
			var color:Object = ColorUtil.hexArgb(mtlSet.ambiantColor);
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT,  3, Vector.<Number>([Number(color.rouge)/255,Number(color.vert)/255,Number(color.bleu)/255,1]));
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function setVertexBufferAt(index:int, vertexBuffer:VertexBuffer3D, bufferOffset:int, format:String):void {
			_context3D.setVertexBufferAt(index, vertexBuffer, bufferOffset, format);
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function setProgramConstantsFromMatrix(type:String,firstRegister:int, matrix:Matrix3D, transposedMatrix:Boolean=false):void {
			_context3D.setProgramConstantsFromMatrix(type, firstRegister, matrix, transposedMatrix);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function setProgramConstantsFromVector(type:String,firstRegister:int, data:Vector.<Number>, numRegisters:int=-1):void {
			_context3D.setProgramConstantsFromVector(type, firstRegister, data, numRegisters);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function destroy():void {
			_shaderStore.destroy();
			_shaderStore = null;
			_parser.destroy();
			_parser = null;
			_mesh.destroy();
			_mesh = null;
			_lights.length = 0;
			_lights = null;
			
			context3D.setVertexBufferAt(0, null);
			context3D.setVertexBufferAt(1, null);
			context3D.setVertexBufferAt(2, null);
			context3D.setVertexBufferAt(3, null);			
			context3D.setVertexBufferAt(4, null);			
			context3D.setVertexBufferAt(5, null);			
			context3D.setVertexBufferAt(6, null);			
			context3D.setVertexBufferAt(7, null);			
			_context3D.setTextureAt(0, null);		
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------
		public function get position():Vector3D {return _modelPosition;	}
		public function set position(value:Vector3D):void {
			_modelPosition.x = value.x;
			_modelPosition.y = value.y;
			_modelPosition.z = value.z;
			
			_modelMatrix.appendTranslation(_modelPosition.x, _modelPosition.y, _modelPosition.z);
		}
		
		public function get modelMatrix():Matrix3D {return _modelMatrix;}
		public function set modelMatrix(value:Matrix3D):void {_modelMatrix = value;}

		public function get context3D():Context3D {return _context3D;}
		public function set context3D(value:Context3D):void {_context3D = _context3D;}	
		
		public function get modelViewProjection():Matrix3D {return _modelViewProjection;}
		public function set modelViewProjection(value:Matrix3D):void {_modelViewProjection = value;}
		
		public function get ambiantLight():Vector.<Number> {return _ambiantLight;}
		public function set ambiantLight(value:Vector.<Number>):void {_ambiantLight = value;}
		
		public function set cameraMatrix(value:Matrix3D):void {	_cameraMatrix = value;	}
		public function get cameraMatrix():Matrix3D {return _cameraMatrix;}		
		
		public function get screenTexture():Texture {return _screenTexture;}
		
		public function set screenTexture(value:Texture):void {_screenTexture = value;	}
		
		public function get lowestHighest():Object { return _parser.highestLowest; }
		
		public function get xTextureScale():Number {return _xTextureScale;}
		public function set xTextureScale(value:Number):void {_xTextureScale = value;}
		
		public function get yTextureScale():Number {return _yTextureScale;}
		public function set yTextureScale(value:Number):void {_yTextureScale = value;}
		
		public function get xOffset():Number {return _xOffset;}
		public function set xOffset(value:Number):void {_xOffset = value;}
		
		public function get xRange():Number {return _xRange;}
		public function set xRange(value:Number):void {	_xRange = value;}
		
		public function get yOffset():Number {return _yOffset;}
		public function set yOffset(value:Number):void {_yOffset = value;}
		
		public function get yRange():Number {return _yRange;}
		public function set yRange(value:Number):void {	_yRange = value;}		
	}	
}