/*Copyright (c) 2012 Moussa Dembélé (www.mousman.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package  
{
	import com.adobe.utils.PerspectiveMatrix3D;
	import com.mousman.as3.agal.arcball.MArcBall;
	import com.mousman.as3.agal.camera.BasicCamera;
	import com.mousman.as3.agal.camera.IMCamera;
	import com.mousman.as3.agal.camera.OrbitalCamera;
	import com.mousman.as3.agal.light.IMLight;
	import com.mousman.as3.agal.light.MLight;
	import com.mousman.as3.agal.mesh.IMMesh;
	import com.mousman.as3.utils.color.ColorUtil;
	import flash.display.Sprite;
	import flash.display.Stage;
	import flash.display3D.Context3D;
	import flash.display3D.Context3DTextureFormat;
	import flash.display3D.Context3DTriangleFace;
	import flash.display3D.textures.Texture;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.geom.Matrix3D;
	import flash.geom.Orientation3D;
	import flash.geom.Point;
	import flash.geom.Vector3D;
	import flash.text.TextField;
	import flash.text.TextFieldAutoSize;
	import flash.text.TextFormat;
	import flash.utils.getTimer;
	/**
	 * ...
	 * @author Moussa Dembélé (www.mousman.com)
	 */
	public class World extends Sprite
	{
		private var _cameras:Vector.<IMCamera>;
		private var _meshes:Vector.<IMMesh>;
		private var _projectionMatrix:PerspectiveMatrix3D;
		private var _modelViewProjection:Matrix3D;
		private var _planViewProjection:Matrix3D;
		
		private var _ambiantLight:Vector.<Number>;
		
		private var _stage:Stage;
		private var _width:int;
		private var _height:int;
		
		private var _context:Context3D;
		
		private var t:Number = 2;
		
		private var fpsLast:uint = getTimer();
		private var fpsTicks:uint = 0;
		private var fpsTf:TextField;	
		private var looptemp:int = 0;		
		private var _mouseDown:Boolean = false;
		
		private var _lights:Vector.<IMLight>;
		
		private var _bgColor:Object;
		
		private var _textureMap:Texture;
		
		private var _arcBall:MArcBall;
		
		private var _lowestHighestPlan:Object;
		
		private var _zZoomFrontier:Number;
		
		private var _originalModelCameraZ:Number;
		
		private var _textMatrix:Matrix3D;
		private var _vecsTextMatrix:Vector.<Vector3D>;
		
		public function World(stage:Stage) 
		{
			// wait for Stage3D to provide us a Context3D
			_stage = stage;
			_width = _stage.stageWidth;
			_height = _stage.stageHeight;
			stage.stage3Ds[0].addEventListener(Event.CONTEXT3D_CREATE, onCreate);
			stage.stage3Ds[0].requestContext3D();			
			
			initGUI();
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function initGUI():void 
		{
			var myFormat:TextFormat = new TextFormat();
			myFormat.color = 0xffffff;
			myFormat.size = 13;
			// create an FPSCounter that displays the framerate on screen
			fpsTf = new TextField();
			fpsTf.x = 0;
			fpsTf.y = 0;
			fpsTf.selectable = false;
			fpsTf.autoSize = TextFieldAutoSize.LEFT;
			fpsTf.defaultTextFormat = myFormat;
			fpsTf.text = "Initializing Stage3d...";
			addChild(fpsTf);			
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function onCreate(e:Event):void 
		{
			removeEventListener(Event.ENTER_FRAME,render);
			_context = _stage.stage3Ds[0].context3D;
			
			if ( _context == null) return;
			
			_context.configureBackBuffer(_width, _height, 4, true);
			_context.setCulling(Context3DTriangleFace.BACK);
			// debug mode
			_context.enableErrorChecking = false;

			var center:Point = new Point(_stage.stageWidth / 2, _stage.stageHeight / 2);
			build();
			_arcBall = new MArcBall(_meshes[1].modelMatrix, center, _stage.stageHeight / 2);
			setArcBallListener();
			
			
			var bgColor:uint = 0x0b0601;
			var decColor:Object = ColorUtil.hexArgb(bgColor);
			_bgColor = { red:decColor.rouge / 255, green:decColor.vert / 255, blue:decColor.bleu / 255};
			
			addEventListener(Event.ENTER_FRAME, render);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function build():void {
			_textureMap = _context.createTexture(1024, 1024, Context3DTextureFormat.BGRA,true);
			_vecsTextMatrix = new Vector.<Vector3D>(3, true);
			_vecsTextMatrix[0] = new Vector3D(1, 1, 1, 1);
			_vecsTextMatrix[2] = new Vector3D(1, 1., 1, 1);
			_textMatrix = new Matrix3D();
			
			_meshes = new Vector.<IMMesh>();
			_modelViewProjection = new Matrix3D();
			setProjection();
			setMeshs();
			setCamera();
			setLights();
			computeZoomFrontier();
			
			_meshes[0].setSaticConstants();
			_meshes[1].setSaticConstants();
			
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function setProjection():void 
		{
			_projectionMatrix = new PerspectiveMatrix3D();
			var aspect:Number = _width /_height; 
			var zNear:Number = 0.1; 
			var zFar:Number = 10000; 
			var fov:Number = 45 * Math.PI / 180; 
			_projectionMatrix.perspectiveFieldOfViewLH(fov, aspect, zNear, zFar);		
		}			
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function setMeshs():void 
		{
			var mesh:DeerMesh = new DeerMesh(_context);
			mesh.position = new Vector3D(0, 0,   0);
			mesh.texMatrix = _textMatrix;
			
			
			
			_meshes.push(mesh);		
			
			var planMesh:PlanMesh = new PlanMesh(_context);
			planMesh.position = new Vector3D(0, 0, 0);
			planMesh.screenTexture = _textureMap;
			_meshes.push(planMesh);
			_lowestHighestPlan = planMesh.lowestHighest;
			
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function setCamera():void {
			_cameras = new Vector.<IMCamera>;
			var cameraPos:Vector3D = new Vector3D(0, 0, -75);
			_originalModelCameraZ = cameraPos.z;
			var basicCamera:BasicCamera = new BasicCamera(cameraPos);
			var orbitalCamera:OrbitalCamera = new OrbitalCamera(_meshes[1].modelMatrix, _stage, 230);
			
			_meshes[0].cameraMatrix = basicCamera.matrix;
			_meshes[1].cameraMatrix = orbitalCamera.matrix;
			_cameras.push(basicCamera);
			_cameras.push(orbitalCamera);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function setLights():void 
		{
			_lights = new Vector.<IMLight>(2,true);
			var light:MLight = new MLight(0xffffff,1.2, 0xffffff, 5);
			light.position = new Vector3D( -50, 8000, -5000);
			// light must be declare to the mesh for shading computation
			_meshes[0].addLight(light);
			
			light = new MLight(0xffffff,1.2, 0xffffff, 5);
			light.position = new Vector3D( -10, 50,-200);
			// light must be declare to the mesh for shading computation
			_meshes[1].addLight(light);
			
			
			_ambiantLight =  Vector.<Number>([1, 1, 1, 0.1]);			
			// light must be declare to the mesh for shading computation
			_meshes[0].ambiantLight  = _ambiantLight;			
			_meshes[1].ambiantLight  = _ambiantLight;			
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function setCameraListener():void 
		{
			// stage listeners for camera control
			_stage.addEventListener(MouseEvent.MOUSE_DOWN, mouseMove);
			_stage.addEventListener(MouseEvent.MOUSE_UP, mouseMove);
			_stage.addEventListener(MouseEvent.MOUSE_MOVE, mouseMove);
			_stage.addEventListener(MouseEvent.MOUSE_WHEEL, mouseMove);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function setArcBallListener():void 
		{
			 //stage listeners for arcBall control
			_stage.addEventListener(MouseEvent.MOUSE_DOWN, mouseMove);
			_stage.addEventListener(MouseEvent.MOUSE_UP, mouseMove);
			_stage.addEventListener(MouseEvent.MOUSE_MOVE, mouseMove);
			_stage.addEventListener(MouseEvent.MOUSE_WHEEL, mouseMove);
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function computeZoomFrontier():void 
		{
			var meshPlan:IMMesh = _meshes[1];
			var zx:Number;
			var zy:Number;
			
			var matrix:Matrix3D = new Matrix3D();
			matrix.append(_projectionMatrix);
			
			var deltax:Number = 178;
			var deltay:Number = 178;
			
			var vec1:Vector3D = new Vector3D(deltax, deltay);
			vec1 = matrix.transformVector(vec1);
			
			zx = vec1.x / 2;
			zy = vec1.y / 2;
			
			_zZoomFrontier = Math.max(zx , zy);
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function mouseMove(evt:MouseEvent ):void 
		{
			switch(evt.type) {
				case MouseEvent.MOUSE_DOWN : 
					_arcBall.mouseDown(); 
					_mouseDown = true;
				break;
				case MouseEvent.MOUSE_UP : 
					_arcBall.mouseUp();  
					_mouseDown = false;
				break;
				case MouseEvent.MOUSE_MOVE : _arcBall.mouseMove(stage.mouseX, stage.mouseY);  break;
				case MouseEvent.MOUSE_WHEEL : _arcBall.mouseWheel(evt.delta);  break;
				case MouseEvent.MOUSE_UP : _arcBall.mouseUp();  break;
				default:break;
			}
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function computeMVMatrix(mesh:IMMesh):Matrix3D
		{
			// clear the matrix and append new angles
			var matrix:Matrix3D = new Matrix3D();
			matrix.append(mesh.modelMatrix);

			var viewMatrix:Matrix3D = mesh.cameraMatrix.clone();
			viewMatrix.invert();
			matrix.append(viewMatrix);
			matrix.append(_projectionMatrix);				
			
			return matrix;
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function render(e:Event ):void 
		{
			if ( !_context ) return;

			var rot:Number;
			var rx:Number;
			var ry:Number;
			var rz:Number;
			var meshModel:IMMesh = _meshes[0];
			var meshPlan:PlanMesh = _meshes[1] as PlanMesh;
			var modelMatrix:Matrix3D;
			var planRation:Number;
			var zoomRatio:Number;
			var modelPosition:Vector3D;
			var matrix:Matrix3D;
			
			
			///////////////////
			 // plan model
			 //////////////////			
			// move plan
			if (!_mouseDown) {
				modelMatrix = meshPlan.modelMatrix;				
				rx = (t/4 * 0.35) % 360 ;
				ry = (t/4 * 0.35) % 360 ;
				rz = (t / 4 * 0.35) % 360 ;
				 //moving plan
				modelMatrix.appendRotation(rx, Vector3D.X_AXIS,modelMatrix.position);
				modelMatrix.appendRotation(ry, Vector3D.Y_AXIS,modelMatrix.position);
				modelMatrix.appendRotation(rz, Vector3D.Z_AXIS,modelMatrix.position);
				
			}
			else {
				// moving plan
				_arcBall.update();
			}
			
			 //compute the modelViewProjection matrix
			meshPlan.modelViewProjection = computeMVMatrix(meshPlan);
			
			// determining zoome ratio
			var deltaZ:Number = meshPlan.modelMatrix.position.z - meshPlan.cameraMatrix.position.z;
			zoomRatio = _zZoomFrontier / deltaZ;
			var cx:Number = meshModel.cameraMatrix.position.x;
			var cy:Number = meshModel.cameraMatrix.position.y;
			var cz:Number = _originalModelCameraZ / zoomRatio;
			meshModel.cameraMatrix.position = new Vector3D(cx,cy,cz);
			
			
			///////////////////
			 // textured model
			 //////////////////
			 modelMatrix = meshModel.modelMatrix;

			// rotating the model
			var axis:Vector3D = new Vector3D(0, 1, 0);
			modelMatrix.appendRotation(2, axis, modelMatrix.position);
			
			// apply plan rotation to model
			var quat:Vector3D = meshPlan.modelMatrix.decompose(Orientation3D.QUATERNION)[1];
			_textMatrix.identity();
			_vecsTextMatrix = _textMatrix.decompose(Orientation3D.QUATERNION);
			_vecsTextMatrix[1] = quat;
			_textMatrix.recompose(_vecsTextMatrix, Orientation3D.QUATERNION);
			
			matrix = new Matrix3D();
			var viewMatrix:Matrix3D = meshModel.cameraMatrix.clone();
			viewMatrix.invert();
			matrix.append(viewMatrix);
			matrix.append(_projectionMatrix);				
			meshModel.modelViewProjection = matrix;
			
			
			// draw model
			  //set render target to texture
			_context.setRenderToTexture(_textureMap,true);
			_context.clear(0.9, 0.9, 0.9, 1);			 
			meshModel.drawTriangles();
			
			// draw plan	
			 //set render target to backbuffer
			_context.setRenderToBackBuffer();
			_context.clear(_bgColor.red, _bgColor.green, _bgColor.blue, 1);					
			meshPlan.drawTriangles();
			
			_context.present();			
			
			// stats
			fpsTicks++;
			var now:uint = getTimer();
			var delta:uint = now - fpsLast;
			// only update the display once a second
			if (delta >= 1000){
				var fps:int = (0.5 + fpsTicks / delta * 1000);
				fpsTf.text = fps + "/60 fps";
				fpsTicks = 0;
				fpsLast = now;
				}				
				addChild(fpsTf);
		}		
		
	}

}